from flask import Flask, render_template
from flask_fas_openid import FAS
from flask.ext.sqlalchemy import SQLAlchemy
from flask.ext.assets import Environment, Bundle
import logging
import logging.handlers
import os

from util.login import FakeFas

# the version as used in setup.py and docs
__version__ = "0.4.4.9"


# Flask App
app = Flask(__name__)

fas = None

# Flask-Assets
assets = Environment(app)
assets.auto_build = False
assets.cache = False
assets.manifest = False

css = Bundle('css/app-foundation.css', 'css/app.css', 'css/jquery.qtip.css', 'css/text.css',
            'css/general_foundicons.css', filters='cssmin', output='css/packed.css')

js = Bundle('js/jquery.js', 'js/jquery.tablesorter.js', 'js/jquery.foundation.buttons.js',
            'js/jquery.foundation.forms.js', 'js/modernizr.foundation.js', 'js/jquery.qtip.js',
            'js/jquery.metadata.js', filters='uglifyjs', output='js/packed.js')

assets.register('css_all', css)
assets.register('js_all', js)


# load default configuration
if os.getenv('DEV') == 'true':
    app.config.from_object('blockerbugs.config.DevelopmentConfig')
    fas = FakeFas(app)
elif os.getenv('TEST') == 'true':
    fas = FAS(app)
    app.config.from_object('blockerbugs.config.TestingConfig')
else:
    fas = FAS(app)
    app.config.from_object('blockerbugs.config.ProductionConfig')


# load real configuration values to override defaults
config_file = '/etc/blockerbugs/settings.py'
if not app.testing:
    if os.path.exists(config_file):
        app.logger.info('Loading configuration from %s' % config_file)
        app.config.from_pyfile(config_file)
    else:
        config_file = os.path.abspath('conf/settings.py')
        if os.path.exists(config_file):
            app.logger.info('Loading configuration from %s' % config_file)
            app.config.from_pyfile(config_file)
        else:
            app.logger.info('No extra config found, using defaults')


# setup logging
fmt = '[%(filename)s:%(lineno)d] ' if app.debug else '%(module)-12s '
fmt += '%(asctime)s %(levelname)-7s %(message)s'
datefmt = '%Y-%m-%d %H:%M:%S'
loglevel = logging.DEBUG if app.debug else logging.INFO
formatter = logging.Formatter(fmt=fmt, datefmt=datefmt)


def setup_logging():
    root_logger = logging.getLogger('')
    root_logger.setLevel(logging.DEBUG)

    if app.config['STREAM_LOGGING']:
        app.logger.debug("doing stream logging")
        stream_handler = logging.StreamHandler()
        stream_handler.setLevel(loglevel)
        stream_handler.setFormatter(formatter)
        root_logger.addHandler(stream_handler)
        app.logger.addHandler(stream_handler)

    if app.config['SYSLOG_LOGGING']:
        app.logger.debug("doing syslog logging")
        syslog_handler = logging.handlers.SysLogHandler(
            address='/dev/log', facility=logging.handlers.SysLogHandler.LOG_LOCAL4)
        syslog_handler.setLevel(loglevel)
        syslog_handler.setFormatter(formatter)
        root_logger.addHandler(syslog_handler)
        app.logger.addHandler(syslog_handler)

    if app.config['FILE_LOGGING'] and app.config['LOGFILE']:
        app.logger.debug("doing file logging to %s" % app.config['LOGFILE'])
        file_handler = logging.handlers.RotatingFileHandler(app.config['LOGFILE'], maxBytes=500000, backupCount=5)
        file_handler.setLevel(loglevel)
        file_handler.setFormatter(formatter)
        root_logger.addHandler(file_handler)
        app.logger.addHandler(file_handler)


setup_logging()

app.logger.debug('using DBURI: %s' % app.config['SQLALCHEMY_DATABASE_URI'])

# database
db = SQLAlchemy(app)


@app.template_filter('tagify')
def tagify(value):
    return value.replace(' ', '')


@app.template_filter('urlsplit')
def urlsplit(value):
    return value.split('/')


@app.template_filter('getname')
def getname(value, entry=1):
    return value.split('/')[entry]


@app.template_filter('getsubname')
def getsubname(value):
    urlparts = value.split('/')
    if len(urlparts) >= 5:
        return ''.join(urlparts[2:4])
    else:
        return ''


@app.template_filter('updatetype')
def updatetype(update):
    is_fe = False
    for bug in update.bugs:
        if bug.proposed_blocker or bug.accepted_blocker:
            return 'blocker'
        elif bug.proposed_fe or bug.accepted_fe:
            is_fe = True
    if is_fe:
        return 'FE'


@app.template_filter('updatelabel')
def updatelabel(bug):
    from models.update import Update
    label = []
    lowest_status_update = bug.updates.filter(Update.status != 'obsolete', Update.status != 'deleted').first()
    if lowest_status_update:
        if lowest_status_update.status == 'stable':
            if bug.status in ['MODIFIED', 'ON_QA', 'VERIFIED', 'CLOSED']:
                label.append('<span class="radius success label">')
                if lowest_status_update.pending:
                    label.append('pending ')
                label.append('stable</span>')

        elif lowest_status_update.status == 'testing':
            label.append('<span class="radius alert label">')
            if lowest_status_update.pending:
                label.append('pending ')
            label.append('testing</span>')

    return ''.join(label)

@app.template_filter('datetime')
def datetime_format(value, format='%Y-%m-%d %H:%M:%S UTC'):
    if value is not None:
        return value.strftime(format)
    return ''


# register blueprints
from blockerbugs.controllers.main import main
app.register_blueprint(main)

from blockerbugs.controllers.admin import admin

from blockerbugs.controllers.api import api_v0
app.register_blueprint(api_v0)

# setup error handling
@app.errorhandler(404)
def page_not_found(e):
    return render_template('404.html', title="Page Not Found!"), 404
