#
# spin.py - database model for spins
#
# Copyright 2012, Red Hat, Inc
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Authors:
#   Tim Flink <tflink@redhat.com>
from datetime import datetime

from sqlalchemy.ext.hybrid import hybrid_property

from blockerbugs import db
from blockerbugs.models import DeclEnum

used_updates = db.Table(
    'used_updates',
    db.Column('update_id', db.Integer, db.ForeignKey('update.id')),
    db.Column('spin_id', db.Integer, db.ForeignKey('spin.id'))
)


class SpinState(DeclEnum):
    requested = 'requested', u'requested'
    built = 'built', u'built'
    obsolete = 'obsolete', u'obsolete'


class SpinType(DeclEnum):
    smoke = 'smoke', u'smoke'
    TC = 'TC', u'TC'
    RC = 'RC', u'RC'

class Spin(db.Model):
    id = db.Column(db.Integer, primary_key=True)
    name = db.Column(db.String(160), unique=False)
    _state = db.Column('state', SpinState.db_type())
    url = db.Column(db.String(2048), nullable=True)
    date_requested = db.Column(db.DateTime, unique=False,
                               default=datetime.utcnow)
    date_created = db.Column(db.DateTime, unique=False)
    _spin_type = db.Column('spin_type', SpinType.db_type())
    updates = db.relationship('Update', secondary=used_updates,
                              backref=db.backref('spins', lazy='dynamic'))
    succeeds_id = db.Column(db.Integer, db.ForeignKey('spin.id'),
                            nullable=True)
    succeeded_by = db.relationship('Spin', backref=db.backref('succeeds',
                                                          remote_side=id,
                                                          uselist=False))
                            #lazy='dynamic')
    milestone_id = db.Column(db.Integer, db.ForeignKey('milestone.id'))
    milestone = db.relationship('Milestone',
                                backref=db.backref('spins', lazy='dynamic'))

    @hybrid_property
    def state(self):
        return self._state

    @state.setter
    def state(self, state):
        if isinstance(state, basestring):
            state = SpinState.from_string(state)
        if self._state == state:
            return
        self._state = state
        if state == SpinState.requested:
            self.date_requested = datetime.utcnow()
        if state == SpinState.built and self.date_created is None:
            self.date_created = datetime.utcnow()

    @hybrid_property
    def spin_type(self):
        return self._spin_type

    @spin_type.setter
    def spin_type(self, spin_type):
        if isinstance(spin_type, basestring):
            spin_type = SpinType.from_string(spin_type)
        self._spin_type = spin_type

    def __str__(self):
        return 'spin: %s' % (self.name)
