#
# bug_sync.py - sync bug information from bugzilla
#
# Copyright 2012, Red Hat, Inc
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Authors:
#   Tim Flink <tflink@redhat.com>

from blockerbugs.models.bug import Bug
from blockerbugs.models.milestone import Milestone
from blockerbugs.util import bz_interface
from datetime import datetime
from sqlalchemy import or_
import re
import logging


class BugSync():
    def __init__(self, db, bzinterface=None, url=None):
        self.db = db
        self.closedre = re.compile('.*CLOSED.*')
        self.log = logging.getLogger('bug_sync')
        if not bzinterface:
            self.bzinterface = bz_interface.BlockerBugs(url=url)
        else:
            self.bzinterface = bzinterface

    def extract_information(self, bug, tracker_type):
        buginfo = {}
        bug_whiteboard = bug.whiteboard

        # get information about accepted/rejected
        buginfo['0day'] = 'Accepted0Day' in bug_whiteboard
        buginfo['prevrel'] = 'AcceptedPreviousRelease' in bug_whiteboard
        buginfo['accepted'] = ('Accepted%s' % tracker_type) in bug_whiteboard
        buginfo['rejected'] = ('Rejected%s' % tracker_type) in bug_whiteboard
        buginfo['proposed'] = not (buginfo['rejected'] or buginfo['accepted'] or
                                   buginfo['prevrel'] or buginfo['0day'])

        # determine whether or not the bug is active (not CLOSED)
        buginfo['active'] = not self.closedre.match(bug.status)

        # basic info
        buginfo['bug_id'] = bug.bug_id
        buginfo['component'] = bug.component
        buginfo['summary'] = bug.summary
        buginfo['status'] = bug.status
        buginfo['url'] = bug.weburl
        buginfo['needinfo'] =  False
        buginfo['needinfo_requestee'] = ''

        # needinfo flag
        if hasattr(bug, 'flags'):
            needinfos = []
            for flag in bug.flags:
                if flag['name'] == 'needinfo' and flag['status'] == '?' and flag['is_active']:
                    buginfo['needinfo'] = True
                    try:
                        needinfos.append(flag['requestee'])
                    except KeyError, e:
                        pass
            buginfo['needinfo_requestee'] = ', '.join(needinfos)

        # according to bugzilla's docs, this is always in UTC
        last_change_time = bug.last_change_time.timetuple()
        buginfo['last_change_time'] = datetime(last_change_time.tm_year,
                                               last_change_time.tm_mon,
                                               last_change_time.tm_mday,
                                               last_change_time.tm_hour,
                                               last_change_time.tm_min,
                                               last_change_time.tm_sec)

        return buginfo

    def get_tracker_id(self, milestone, tracker_type):
        if tracker_type == 'Blocker':
            return milestone.blocker_tracker
        elif tracker_type == 'FreezeException':
            return milestone.fe_tracker

    def update_milestone(self, milestone, tracker_type, lastupdate=None):
        trackerid = self.get_tracker_id(milestone, tracker_type)

        self.log.info('Updating %s for Fedora %d %s (%d)' % (tracker_type,
                                                             milestone.release.number,
                                                             milestone.version,
                                                             trackerid))

        newbugs = self.bzinterface.do_query(trackerid, last_update=lastupdate)
        self.log.debug('New bugs: %s' % str(newbugs))
        added_bugs = []
        for newbug in newbugs:
            oldbug = Bug.query.filter_by(bugid=newbug.bug_id,
                                         milestone=milestone).first()
            newbuginfo = self.extract_information(newbug, tracker_type)
            if oldbug:
                self.log.debug('Updating bug %d' % newbuginfo['bug_id'])
                oldbug.update(newbuginfo, tracker_type, milestone)
            else:
                self.log.debug('Adding new bug %d' % newbuginfo['bug_id'])
                oldbug = Bug.from_data(newbuginfo, milestone, tracker_type)
                added_bugs.append(oldbug)

            self.db.session.add(oldbug)
            self.db.session.commit()


    def cleanup_milestone(self, milestone, tracker_type):
        trackerid = self.get_tracker_id(milestone, tracker_type)

        self.log.info('Cleaning up  %s for Fedora %d %s (%d)' % (
            tracker_type, milestone.release.number, milestone.version, trackerid))

        tracker_depends = set(self.bzinterface.get_deps(trackerid))
        self.log.debug("Current dependencies for %d: %s" % (trackerid, str(tracker_depends)))
        if tracker_type == 'FreezeException':
            blockers = Bug.query.filter_by(milestone=milestone).filter(or_(Bug.proposed_fe == True,
                                                                    Bug.accepted_fe == True)).all()
        elif tracker_type == 'Blocker':
            blockers = Bug.query.filter_by(milestone=milestone).filter(or_(Bug.proposed_blocker == True,
                                                                    Bug.accepted_blocker == True,
                                                                    Bug.accepted_0day == True,
                                                                    Bug.accepted_prevrel == True)).all()
        for blocker in blockers:
            if not blocker.bugid in tracker_depends:
                if tracker_type == 'FreezeException':
                    blocker.proposed_fe = False
                    blocker.accepted_fe = False
                    blocker.rejected_fe = True
                elif tracker_type == 'Blocker':
                    blocker.proposed_blocker = False
                    blocker.accepted_blocker = False
                    blocker.accepted_0day = False
                    blocker.accepted_prevrel = False
                    blocker.rejected_blocker = True
                self.db.session.add(blocker)
            self.db.session.commit()


    def update_active(self, full_sync=False):
        starttime = datetime.utcnow()
        active_milestones = Milestone.query.filter_by(active=True).all()
        self.log.info("Starting blocker sync")

        # for each active milestone, update the tracker to remove any bugs that
        # no longer block tracker and pull in any new bugs
        for milestone in active_milestones:
            last_sync_time = milestone.last_bug_sync
            if full_sync:
                last_sync_time = None

            self.log.info("Syncing blockers for milestone %s" % milestone)
            self.cleanup_milestone(milestone, 'Blocker')
            self.update_milestone(milestone, 'Blocker', last_sync_time)

            self.log.info("Syncing FE for milestone %s" % milestone)
            self.cleanup_milestone(milestone, 'FreezeException')
            self.update_milestone(milestone, 'FreezeException', last_sync_time)

            # update the last query time for the milestone
            milestone.last_bug_sync = starttime
            self.db.session.add(milestone)
            self.db.session.commit()


