#
# bz_interface.py - common methods for obtaining blocker/fe information from
#                   RHBZ
#
# Copyright 2012, Red Hat, Inc
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Authors:
#   Tim Flink <tflink@redhat.com>


import bugzilla
import logging
import datetime
from xmlrpclib import Fault

base_query = {'o1': 'anywords',
            'f1': 'blocked',
            'query_format':'advanced'}

class BZInterfaceError(Exception):
    def __init__(self, msg):
        self.msg = msg

    def __str__(self):
        return repr(self.msg)


class BlockerBugs():
    def __init__(self, user=None, password=None, url=None, bz=None, logger=None):
        self.logger = logger if logger else logging.getLogger('bz_interface')
        if not bz:
            if not url:
                url = 'https://bugzilla.redhat.com/xmlrpc.cgi'
            if not (user and password):
                self.bz = bugzilla.RHBugzilla4(url=url,
                                               cookiefile=None,
                                               tokenfile=None)
            else:
                self.bz = bugzilla.RHBugzilla4(url=url,
                                               user=user,
                                               password=password,
                                               cookiefile=None,
                                               tokenfile=None)
        else:
            self.bz = bz
        self.logger.info('Using bugzilla URL: %s' % url)
    #https://partner-bugzilla.redhat.com/buglist.cgi?bug_status=NEW&bug_status=ASSIGNED
    #&bug_status=POST&bug_status=MODIFIED&classification=Fedora&component=anaconda&f1=component&o1=changedafter&product=Fedora&query_format=advanced&v1=2013-03-21%2012%3A25&version=19
    def get_bz_query(self, tracker, last_update=None):
        query = {}
        query.update(base_query)
        query['v1'] = str(tracker)
        if last_update:
            last_update_string = last_update.strftime("%Y-%m-%d %H:%M GMT")
            yesterday_timestring = (last_update - datetime.timedelta(days=1)).strftime("%Y-%m-%d")

            # Since this is kind of confusing, the idea here is to only grab only
            # the bugs whose whiteboard, blocks or dependson has changed since
            # the last sync or any bugs that have been created in the last day.

            # Unfortunately, there doesn't seem to be a way to filter the created
            # bugs in a way that's more granular than a day, so we'll pull in
            # extra bugs and slow down the sync process but the alternative is
            # missing bugs from the list

            # keep in mind that OP -> open parenthesis, CP -> closed parenthesis
            # and the OR is to make sure that a bug is returned if ANY of the
            # changed after conditions are met
            query.update({
                'f2': 'OP',
                'j2': 'OR',
                'f3': 'blocked',
                'o3': 'changedafter',
                'v3': last_update_string,
                'f4': 'status_whiteboard',
                'o4': 'changedafter',
                'v4': last_update_string,
                'f5': 'bug_status',
                'o5': 'changedafter',
                'v5': last_update_string,
                'f6': 'dependson',
                'o6': 'changedafter',
                'v6': last_update_string,
                'f7': 'creation_ts',
                'o7': 'greaterthaneq',
                'v7': yesterday_timestring,
                'f8': 'short_desc',
                'o8': 'changedafter',
                'v8': last_update_string,
                'f9': 'component',
                'o9': 'changedafter',
                'v9': last_update_string,
                'f10': 'CP'
            })
        return query


    def do_query(self, tracker, last_update=None):
        query = self.get_bz_query(tracker, last_update)
        buglist = self.flatten_bug_list(query)
        self.logger.debug('Doing query: %s' % str(query))
        return buglist


    # instead of having a bunch of nested bug objects, make a final list of
    # all of the applicable bugs
    def flatten_bug_list(self, query, deps=None, max_recurse=3):
        result = self.bz.query(query)

        if max_recurse <= 0:
            return result

        if deps == None:
            deps = []

        for i in range(len(result)):
            b = result[i]
            if len(b.dependson) > 0:
                for dep in b.dependson:
                    if dep not in deps:
                        deps.append(dep)
                        result.extend(self.flatten_bug_list(self.get_bz_query(dep),
                                                            max_recurse=max_recurse - 1,
                                                            deps=deps))

        return result

    def get_deps(self, bugid):
        return self.bz.getbug(bugid).dependson


class BlockerProposal():
    def __init__(self, bz, bugid, trackers, is_blocker=False, is_fe=False, bz_user=''):
        self.bz = bz
        self.bugid = bugid
        self.trackers = trackers
        self.is_blocker = is_blocker
        self.is_fe = is_fe
        self.bz_user = bz_user
        self.bugid_ok = False
        self.proposal_ok = False
        self.bugdata = None
        self.blocks = None

        self.log = logging.getLogger('bz_interface.bugproposal')

    def get_bugdata(self):
        # get bug data, will raise XMLRPC fault if the bug does not exist
        try:
            self.bugdata = self.bz.getbug(self.bugid)
        except Fault, e:
            if e.faultCode == 101:
                raise BZInterfaceError(e.faultString)
            else:
                raise

    def get_tracker_type(self):
        if self.is_blocker and self.is_fe:
            return 'Blocker and Freeze Exception'
        if self.is_blocker:
            return 'Blocker'
        if self.is_fe:
            return 'Freeze Exception'

    def propose_bugs(self, bz_user, milestone_name, justification):
        comment = ['Proposed as a', self.get_tracker_type(), 'for', milestone_name, 'by',
                   bz_user, 'using the blocker tracking app because:\n\n',
                   justification]
        tracker_bugs = []
        if self.is_blocker:
            tracker_bugs.append(self.trackers['blocker'])
        if self.is_fe:
            tracker_bugs.append(self.trackers['fe'])
        self.log.info('comment: %s' % comment)
        try:
            self._do_proposal(tracker_bugs, self.bugid, ' '.join(comment), self.bz_user)
        except Fault, e:
            if e.faultCode == 51:
                # bugzilla account does not exist, this should happen very rarely, if ever
                # so just redo the call with nothing to add to cc
                # TODO - it might be useful to ask the user to re-associate accounts again here
                self._do_proposal(tracker_bugs, self.bugid, ' '.join(comment), '')
            else:
                raise BZInterfaceError(e.faultString)

    def _do_proposal(self, tracker, proposed_bugid, comment, bz_user):
        bug_update = self.bz.build_update(blocks_add=tracker,
                                          cc_add=[bz_user],
                                          comment=comment)
        self.bz.update_bugs(proposed_bugid, bug_update)

    def check_proposed_bug(self):
        if not self.bugdata:
            self.get_bugdata()

        # check to make sure that bug is not already CLOSED
        bug_status = self.bugdata.bug_status
        if 'CLOSED' in bug_status.split():
            raise BZInterfaceError('Bug %i is CLOSED: %s' % (self.bugid, bug_status))

    def get_blocks(self):
        if not self.bugdata:
            self.get_bugdata()
        self.blocks = self.bugdata.blocked

    def check_blocker_proposal(self):
        if not self.blocks:
            self.get_blocks()

        if self.trackers['blocker'] in self.blocks:
            return False
        return True

    def check_fe_proposal(self):
        if not self.blocks:
            self.get_blocks()

        if self.trackers['fe'] in self.blocks:
            return False
        return True

