#
# update_sync.py - sync update information with bodhi
#
# Copyright 2012, Red Hat, Inc
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Authors:
#   Tim Flink <tflink@redhat.com>
from datetime import datetime
import logging

from fedora.client import bodhi, ServerError

from blockerbugs import app
from blockerbugs.models.update import Update
from blockerbugs.models.bug import Bug

bodhi_baseurl = app.config['BODHI_URL']


class UpdateSync(object):
    def __init__(self, db, bodhi_interface=None):
        self.db = db
        if bodhi_interface:
            self.bodhi_interface = bodhi_interface()
        else:
            self.bodhi_interface = bodhi.BodhiClient(base_url=bodhi_baseurl)
        self.log = logging.getLogger('update_sync')

    def extract_information(self, update):
        updateinfo = {}
        date_pushed = None
        if update.date_pushed:
            date_pushed = datetime.strptime(update.date_pushed,
                                            '%Y-%m-%d %H:%M:%S')
        updateinfo['date_pushed_testing'] = None
        updateinfo['date_pushed_stable'] = None
        updateinfo['pending'] = False
        if update.status == 'pending':
            updateinfo['pending'] = True

            if update.request == 'stable':
                updateinfo['status'] = 'stable'
                updateinfo['date_pushed_testing'] = date_pushed
            elif update.request == 'testing':
                updateinfo['status'] = 'testing'
            else:
                updateinfo['status'] = 'undefined'
        else:
            updateinfo['status'] = str(update.status)
            if update.status == 'testing':
                updateinfo['date_pushed_testing'] = date_pushed
            elif update.status == 'stable':
                updateinfo['date_pushed_stable'] = date_pushed

        updateinfo['title'] = str(update.title)
        updateinfo['karma'] = update.karma
        updateinfo['stable_karma'] = update.stable_karma

        if update.updateid is None:
            urlsufix = updateinfo['title']
        else:
            urlsufix = str(update.updateid)
        updateinfo['url'] = '%s%s' % (bodhi_baseurl, urlsufix)

        updateinfo['date_submitted'] = datetime.strptime(update.date_submitted,
                                                         '%Y-%m-%d %H:%M:%S')
        updateinfo['bugs'] = []
        if len(update.bugs) > 0:
            for buginfo in update.bugs:
                updateinfo['bugs'].append(buginfo.bug_id)

        return updateinfo

    def get_update(self, envr):
        updates = self.bodhi_interface.query(package=envr)
        return self.extract_information(updates['updates'][0])

    def is_watched_bug(self, bugnum):
        watched_bug = Bug.query.filter_by(bugid=bugnum).first()
        if watched_bug:
            return True
        else:
            return False

    def clean_updates(self, updates):
        query_updates = Update.query.filter(Update.status != 'obsolete', Update.status != 'deleted').all()
        db_updates = set(update.title for update in query_updates)
        bodhi_updates = set(update['title'] for update in updates)
        deleted_updates = db_updates.difference(bodhi_updates)

        for update in query_updates:
            if update.title in deleted_updates:
                update.status = 'deleted'
                self.db.session.add(update)
                self.db.session.commit()

    def search_updates(self, bugids, release_num):
        queries_data = dict(
                bugs=[str(bug_id) for bug_id in bugids],
                release='f%d' % release_num,
                limit=100,
        )
        try:
            result = self.bodhi_interface.query(**queries_data)
        except bodhi.BodhiClientException, e:
            # ignore invalid release error so it doesn't blow up when
            # bug trackers for a release not present in bodhi are added
            if 'Invalid release' in e.message:
                self.log.exception(e)
                return []
            else:
                raise

        if u'status' in result.keys():
            raise ServerError('', 200, result['errors'][0].description)

        updates = {}
        for update in result.updates:
            updates[update.title] = update

        while result.page < result.pages:
            result = self.bodhi_interface.query(page=result.page + 1,
                                                **queries_data)
            for update in result.updates:
                updates[update.title] = update

        updates = updates.values()  # updates without duplicates
        self.log.info('found %d updates from bodhi for %d bugs in f%d' %
                      (len(updates), len(bugids), release_num))
        return [self.extract_information(update) for update in updates]

    def get_release_bugs(self, release):
        buglist = []
        for milestone in release.milestones:
            buglist.extend(milestone.bugs.filter_by(active=True).all())
        return buglist

    def sync_bug_updates(self, release, bugs, fullsync=False):
        starttime = datetime.utcnow()
        bugs_ids = [bug.bugid for bug in bugs]
        self.log.debug('searching for updates for bugs %s' % str(bugs_ids))
        try:
            updates = self.search_updates(bugs_ids, release.number)
        except ServerError, ex:
            self.log.error(
                'f{r.number} sync updates failed: {e.code} {e.msg}'.format(
                    e=ex, r=release))
            return

        # mark deleted updates
        self.clean_updates(updates)

        for update in updates:
            self.log.debug('running sync for update %s' % update['title'])
            existing_update = Update.query.filter_by(
                title=update['title']).first()
            if existing_update:
                self.log.debug(
                    'syncing existing update %s' % existing_update.title)
                existing_update.sync(update, fullsync)
            else:
                self.log.debug('creating new update %s' % update['title'])
                existing_update = Update.from_data(update, release)
            self.db.session.add(existing_update)
            self.db.session.commit()

        release.last_update_sync = starttime
        self.db.session.add(release)
        self.db.session.commit()

    def sync_updates(self, release, fullsync=False):
        bugs = self.get_release_bugs(release)
        self.log.info(
            'found %d bugs in f%d release' % (len(bugs), release.number))
        if not bugs:
            self.log.info(
                'no bugs in f%d release, skip update' % (release.number))
            release.last_update_sync = datetime.utcnow()
            return
        self.sync_bug_updates(release, bugs, fullsync)
